/**
 * I want a (non-binary) Tree for use e.g. in TreeViewer + TreeEditor
 * Let's start with an immutable API, bc that is an important pillar of computational / fp goodness
 *
 * a Tree is just a Node
 *
 * Nodes are/have ...
 *    an property key: K ,
 *    an property data: D ,
 *    asympt-runtime-reducing meta info, like
 *      max depth of child trees
 *      own depth
 *      #of subnodes
 *      ... something for specific orderings?
 *
 * ? what data struct shall hold the children?
 * rn: Hashmap with... roughly constant access via key?
 *
 * stretch-goal: let it be cut at depth x
 * either removing nodes below, or squashing them (in _some_ order) into depth x
 */
export type HashMap<K, D> = Map<K, D>;
export type TNodeOrLeaf<K, A> = {
    kind: "leaf";
} | TNode<K, A>;
export type MbyTree<K, A> = TNodeOrLeaf<K, A>;
export type Path<K> = K[];
export declare const TNode_with_children: <K, D>(key: K) => (data: D) => (child_array: TNode<K, D>[]) => TNode<K, D>;
export declare class TNode<K, D> {
    id: K;
    data: D;
    size: bigint;
    children: HashMap<K, TNode<K, D>>;
    max_depth: number;
    constructor(key: K, data: D);
}
export type Tree<K, D> = TNode<K, D>;
export type AutoIndTree<D> = TNode<bigint, D>;
export type AutoIndNode<D> = AutoIndTree<D>;
/**
 *
 * Constructor for a AutoIndTree<D> node
 *
 * @param data
 * @param child_array
 * @returns
 */
export declare const TNode_with_children_autoind: <D>(data: D) => (child_array: AutoIndNode<D>[]) => AutoIndTree<D>;
export declare const append_toplevel: <K, D>(tree: Tree<K, D>) => (node: TNode<K, D>) => Tree<K, D>;
export declare function tNode_no_k<D>(d: D): TNode<bigint, D>;
export declare function tNode_new_k<K, D>(k: K, n: TNode<any, D>): {
    k: K;
    id: any;
    data: D;
    size: bigint;
    children: HashMap<any, TNode<any, D>>;
    max_depth: number;
};
/**
 *
 * @param tree that shall get new child
 * @param node to append - key gets scrubbed and auto assigned
 * @returns
 */
export declare const append_toplevel_autoind: <D>(tree: AutoIndTree<D>) => (n: AutoIndNode<D>) => AutoIndTree<D>;
/**
 * TODO: fix that this function mutates t!
 * @param t
 * @returns
 */
export declare const append_all_toplevel_autoind: <D>(t: AutoIndTree<D>) => (ns: AutoIndNode<D>[]) => AutoIndTree<D>;
export declare const contains: <K, D>(t: Tree<K, D>) => (p: Path<K>) => boolean;
