// import { List } from 'immutable';
//import { Dictionary } from 'typescript-collections';
// import { forEach } from 'typescript-collections/dist/lib/arrays';
// todo: get lodash/fp import to work
// import * as _ from 'lodash';

// todo: squash bug with "always 0n" indexes
//   in append_all_toplevel_autoind or TNode_with_children_autoind?
//   PortfolioComponent.test_tree shows the wrong behavior
/**
 * I want a (non-binary) Tree for use e.g. in TreeViewer + TreeEditor
 * Let's start with an immutable API, bc that is an important pillar of computational / fp goodness
 *
 * a Tree is just a Node
 *
 * Nodes are/have ...
 *    an property key: K ,
 *    an property data: D ,
 *    asympt-runtime-reducing meta info, like
 *      max depth of child trees
 *      own depth
 *      #of subnodes
 *      ... something for specific orderings?
 *
 * ? what data struct shall hold the children?
 * rn: Hashmap with... roughly constant access via key?
 *
 * stretch-goal: let it be cut at depth x
 * either removing nodes below, or squashing them (in _some_ order) into depth x
 */

export type HashMap<K, D> = Map<K, D>;

export type TNodeOrLeaf<K, A> =
  | { kind: "leaf" }
  | TNode<K, A>

export type MbyTree<K, A> = TNodeOrLeaf<K, A>

export type Path<K> = K[] // ? = List<K>

// In retrospect, this is a hilarious/ingenious way to do the type signature.
// Total throwback to my time spent as a Haskell/Functional Programming tutor.
// I don't know how users will feel about the (staggered)(access)(syntax)
// but it might just be nice ^^
// sidenote: it's not the best for memory consumption at runtime
//           since all these intermediary functions are around in some form then,
//           but I think it shouldn't slow down the program much at all
export const TNode_with_children = <K, D>
  (key: K) => (data: D) => (child_array: TNode<K, D>[]): TNode<K, D> => {
    let res = new TNode<K, D>(key, data);
    child_array.forEach((child: TNode<K, D>) => res.children.set(child.id, child));
    Object.freeze(res);
    return res;
  }

export class TNode<K, D> {
  //kind: "non_leaf" | "leaf" = "non_leaf";
  id: K;
  data: D;
  size: bigint = 1n;
  children: HashMap<K, TNode<K, D>> = new Map<K, TNode<K, D>>();
  max_depth: number = 0;

  constructor(key: K, data: D) {
    this.id = key
    this.data = data;
  }
}

export type Tree<K, D> = TNode<K, D>
export type AutoIndTree<D> = TNode<bigint, D>
// TODO ^
export type AutoIndNode<D> = AutoIndTree<D>

/**
 *
 * Constructor for a AutoIndTree<D> node
 *
 * @param data
 * @param child_array
 * @returns
 */
export const TNode_with_children_autoind = <D>
  (data: D) => (child_array: AutoIndNode<D>[]): AutoIndTree<D> => {
    let res = new TNode<bigint, D>(0n, data);
    child_array.forEach((child_node: AutoIndNode<D>, ind: number) => res.children.set(BigInt(ind), tNode_new_k(BigInt(ind), child_node)))
    return res;
  }

export const append_toplevel = <K, D>
  (tree: Tree<K, D>) => (node: TNode<K, D>): Tree<K, D> => {
    let new_child = tree.children
    new_child.set(node.id, node)

    return {
      ...tree,
      children: new_child,
      max_depth: tree.max_depth == 0 ? 1 : tree.max_depth,
      size: tree.size + 1n,
      // todo: update all metadata
      // like max_depth = max previous n.max_depth+1
    }
  }

export function tNode_no_k<D>(d: D) {
  return new TNode<bigint, D>(0n, d)
}

export function tNode_new_k<K, D>(k: K, n: TNode<any, D>) {
  return { ...n, k: k }
}

/**
 *
 * @param tree that shall get new child
 * @param node to append - key gets scrubbed and auto assigned
 * @returns
 */
export const append_toplevel_autoind = <D>
  (tree: AutoIndTree<D>) => (n: AutoIndNode<D>): AutoIndTree<D> => {
    let new_child_hm = tree.children
    let new_ind = BigInt(tree.children.size)
    new_child_hm.set(new_ind, { ...n, id: new_ind }) // TODO: test

    return {
      ...tree,
      children: new_child_hm,
      // todo: update all metadata
      // (copy from above when ready)
    }
  }

/**
 * TODO: fix that this function mutates t!
 * @param t
 * @returns
 */
export const append_all_toplevel_autoind = <D>
  (t: AutoIndTree<D>) => (ns: AutoIndNode<D>[]): AutoIndTree<D> => {
    let first_new_ind = BigInt(t.children.size)
    ns.forEach((appendee, arr_i) => {
      let new_child = tNode_new_k(first_new_ind + BigInt(arr_i), appendee)
      t.children.set(new_child.id, new_child)
    })

    return t
  }


export const contains = <K, D>(t: Tree<K, D>) => (p: Path<K>): boolean => {
  // todo
  return false;
}

// EXAMPLE / TEST OBJECTS
const test_tree_1_3: Tree<bigint, string> =
  TNode_with_children<bigint, string>(0n)("root of 1-3 tree")(
    [new TNode(0n, "child node #1"),
    new TNode(1n, "the label appearing in a path"),
    new TNode(2n, "or a whole task in Todoist"),
    ])

const test_tree_1_4: Tree<bigint, string> =
  append_toplevel_autoind(test_tree_1_3)(new TNode(2n, "2nd root?"))
